class WebSocketManager {
  constructor(url, app, join_token) {
    this.url = url;
    this.params = ''
    this.socketOpen = false;
    this.reconnectInterval = 10000; // 重连间隔时间  
    this.heartbeatInterval = 15000; // 心跳检测间隔时间  
    this.heartbeatTimeout = null; // 心跳检测超时定时器  
    this.reconnectTimer = null; // 重连定时器  
    this.socket = null;
    this.invocationId = 0;
    this.app = app;
    this._join_token = join_token;
    this.global_is_linking = false;
    this.msglist = []; //历史消息
    this.addcount = 0; //连接次数统计
    // this.lastMsgID = 0; //最后一条消息的ID
    this.chat_min_id = 0; //最小消息ID
    this.chat_max_id = 0; //最大消息ID
    this.can_request_pre_chat_page = false; //是否可以获取上一页数据
    this.read_last_page=false;//是否有读取最后一页数据
    this.basicinfo={};
    if (this._join_token) {
      this.initWebSocket();
    }
    this.create_time = new Date();
  }
/**
 * 初始化链接，创建链接 服务端是signalr
 */
  initWebSocket() {
    setTimeout(() => {
      if (this.app.globalData.socket) {
        if (this.app.globalData.socket.readyState == 1) {
          this.socket = this.app.globalData.socket;
          if(this.msglist.length==0){
            console.log('history msglsit ',this.msglist.length);
            this.can_request_pre_chat_page=true;
            this.chat_min_id=0;
            this.chat_max_id=0;
            this.privateGetSocketPrePageMessage();
          }else{
            console.log('history msglsit ',this.msglist.length);
          }
          return;
        }
      }

      if (!this.global_is_linking) {
        this.global_is_linking = true;
        if (!this._join_token) {
          this.global_is_linking = false;
          // console.log('set global_is_linking false from jointoken is empty');
          return console.log('this._join_token为空');
        }
        if (this.reconnectTimer) {
          clearTimeout(this.reconnectTimer);
        }
        if (this.socket) {
          try {
            this.socket = null
            this.socket.close();
            delete this.socket;
          } catch { }
        }
        // console.log('开始链接',this.global_is_linking,new Date());
        wx.request({
          url: `${this.url.replace(/wss:/g, 'https:')}/chathub/negotiate?negotiateVersion=1`, //仅为示例，并非真实的接口地址
          data: {},
          method: "POST",
          header: {
            'content-type': 'application/json', // 默认值
            'appid': wx.getStorageSync('appidwx'),
            'utoken': wx.getStorageSync('userToken')
          },
          success: (rest) => {
            if (rest.data && !rest.data.code) {
              this.params = `?id=${rest.data.connectionToken}`
              this.socket = wx.connectSocket({
                url: `${this.url.replace(/https:/g, 'wss:')}/chathub${this.params}`,
                success: (res) => {
                  console.log('WebSocket连接成功', res);
                },
                fail: (err) => {
                  console.error('WebSocket连接失败', err);
                  this.reconnect();
                }
              });
              this.socket.onOpen((open) => {
                this.addcount++
                this.socketOpen = true;
                this.startHeartbeat();
                this.sendMessage({
                  "protocol": "json",
                  "version": 1
                })
                this.sendJoin()
                setTimeout(() => {
                  this.app.emit('linkSocket', this.socketOpen)
                }, 100)
              })
              this.socket.onMessage((message) => {
                var _data = message.data;
                if (_data) {
                  //处理粘包问题
                  for (var _msg of _data.split(String.fromCharCode(0x1e))) {
                    if (_msg) {
                      this.handlerSocketMessage(_msg);
                    }
                  }
                }
              });
              this.socket.onClose((res) => {
                this.global_is_linking = false;
                console.error('WebSocket已关闭！', res);
                this.socketOpen = false;
                this.app.emit('linkSocket', this.socketOpen)
                this.reconnect();
              });
              this.socket.onError((err) => {
                this.global_is_linking = false;
                console.error('WebSocket错误：', err);
                this.socketOpen = false;
                this.app.emit('linkSocket', this.socketOpen)
                this.reconnect();
              });
              this.app.globalData.socket = this.socket;
            }
          },
          fail: () => {
            this.global_is_linking = false;
            // console.log('set global_is_linking false from api fail');
            this.reconnect()
          }
        })
      } else {
        console.warn(this.global_is_linking, "已经在链接了，无法重复链接");
      }
    }, 100);
  }

  /**
   * 新的链接
   * @param {*} _url 
   * @param {*} _token 
   */
  newLinkConfig(_url,_token){
    if(this._join_token!= _token){
      if(this.socket){
        this.socket.close();
        delete this.socket;
      }
      this.url = _url;
      this._join_token = _token;
      this.global_is_linking=false;
      this.msglist=[];
      this.chat_max_id=0;
      this.chat_min_id=0;
      this.initWebSocket();
    }
  }
  /**
   * 发送当前客户信息，注意plat_from
   * @param {*} data 
   */
  sendJoin(data) {
    // let userinfo = wx.getStorageSync('userInfo')
    let userinfo = null;//这个时三方信息，需要是object对象，可以转化成Json的，plat_from是一个JsonString
    var _info = {
      visitor_token: wx.getStorageSync('visitor_token'),
      join_token: this._join_token,
      plat_from: userinfo ? JSON.stringify(userinfo) : null,
      plat: "wxapp",
      url: data?.url,
      referrer: data ? data.referrer.toString() : '',
      title: data?.title
    };
    this.sendMessage(_info, 'Join')
  }
  /**
   * 重新链接后获取断层消息
   */
  handlerGetSocketFaultMessage() {
    this.sendMessage([this.chat_max_id], 'News');
  }
  /**
   * 下拉，获取上一页消息
   */
  handlerGetSocketPrePageMessage(_minid) {
    var _size = 10;
    if (this.msglist) {
      var _len = this.msglist.length;
      if (_len > 0) {
        var _c_minid = -1;
        var _targetlist = null;
        if (_minid) {
          for (var k = 0; k < this.msglist.length; k++) {
            if (this.msglist[k].id >= _minid) {
              _c_minid = k;
              break;
            }
          }
          if (_c_minid >= 0) {
            var _start = _c_minid - 10;
            if (_start < 0) {
              _start = 0;
            }
            _targetlist = this.msglist.slice(_start, _c_minid);
          }
        } else {
          _targetlist = this.msglist.slice(-10);
          // console.log('从', this.msglist.length, "中获取最后10条");
        }
        if (_targetlist != null && _targetlist.length > 0) {
          _targetlist.reverse(); //升序改成倒叙
          var _strS = JSON.stringify(_targetlist);
          //推送出去
          this.handlerShareMessage('Msgs', JSON.parse(_strS));
          if (_targetlist.length < 10) {
            this.privateGetSocketPrePageMessage();
          }
        } else {
          this.privateGetSocketPrePageMessage();
        }
      } else {
        this.privateGetSocketPrePageMessage();
      }
    }
  }
  /**
   * 下拉，获取上一页消息
   */
  privateGetSocketPrePageMessage() {
    if (this.can_request_pre_chat_page) {
      this.sendMessage([this.chat_min_id, 10], 'Msgs');
    } else {
      //通知UI，没有更早的数据
      this.handlerShareMessage('nodata', false);
    }
  }
  /**
   * 获取最后一页数据 10条 被app引用
   */
  handlerGetLastPageMessage() {
    if(this.can_request_pre_chat_page && !this.read_last_page){
      return null;
    }
    //是否有读取过最后一页，如果没有则返回null 如果是新户，则忽略
      return this.msglist.slice(-10);
  }
  /**
   * 发送单条消息给UI
   * @param {*} _chat  消息chatModel格式
   * @param {*} _append 这个参数作废了
   * @param {*} _from_target 来自哪个Target
   */
  handlerShareChatMessage(_chat, _append, _from_target) {
    var _talkMessageModel = {
      target: 'Msg',
      append: _append,
      from: _from_target || 'Msg',
      data: _chat
    };
    this.app.emit('talkmessage', _talkMessageModel);
  }

  /**
   * 推送消息给UI 注意有格式的
   * @param {*} _target Msg,OpenWindow,OpenScore,OpenMessage等
   * @param {*} _data ChatModel,''
   */
  handlerShareMessage(_target, _data) {
    var _talkMessageModel = {
      target: _target,
      append: true,
      data: _data
    };
    this.app.emit('talkmessage', _talkMessageModel);
  }
  /**
   * 组合消息格式，然后发送出去，注意参数的动态化
   * @param {*} data 
   * @param {*} type 
   */
  sendMessage(data, type) {
    if (this.socket && this.socket.readyState == 1) {
      let saveData = {}
      this.invocationId++
      if (type) {
        saveData = {
          "arguments": Array.isArray(data) ? data : [data],
          "invocationId": this.invocationId.toString(),
          "target": type,
          "type": 1
        }
      } else {
        saveData = data
      }
      this.socket.send({
        data: JSON.stringify(saveData) + String.fromCharCode(0x1e),
        success: () => {
        },
        fail: (err) => {
          console.error('消息发送失败', err);
        }
      });
    } else {
      // if(this.socket){
      //   console.log(this.socket.readyState);
      // }
      //链接没有准备好，是否需要暂存消息
      // if (this._join_token) {
      //   if (type != 'Join' && type != 'Msgs' && type != 'TapWindow') {
      //     //   wx.showToast({
      //     //     title: '服务繁忙，稍后重试',
      //     //     icon: 'none'
      //     //   })
      //   }
      // }
      console.warn('WebSocket未连接，消息发送失败', data);
    }
  }
  /**
   * 处理socket下来的一条消息，综合格式
   * @param {*} message 
   */
  handlerSocketMessage(message) {
    try {
      const data = JSON.parse(message);
      if (data && data.type) {
        switch (data.type) {
          case 1: //下发的消息
            {
              if (data.target) {
                switch (data.target) {
                  case "Msg": {
                    let _chat = JSON.parse(data.arguments[0])
                    if (_chat) {
                      if (_chat.id) {
                        this.handlerSetMaxMinChatId(_chat.id);
                      }
                      this.msglist.push(_chat);
                      this.handlerShareChatMessage(_chat, true);
                    }
                    // this.lastMsgID = _chat.id
                    // this.msglist.push(_chat)
                  }
                    break;
                  case "OpenWindow":
                  case "OpenScore":
                  case "OpenMessage": {
                    this.handlerShareMessage(data.target, '');
                  }
                    break;
                  case "CallBack": {
                    this.handlerShareMessage(data.target, data.arguments[0]);
                  }
                    break;
                  case "Health": {
                    this.sendMessage(wx.getStorageSync('visitor_token') || '', 'Health')
                  }
                    break;
                  case "TapWindow":
                  case "TakeNotify":
                  case "EnterTalk":
                    break;
                  default:
                    console.log('unhandler msg ', data.target, "data.type==1");
                    break;
                }
              }
            }
            break;
          case 3: //回复的消息
            {
              if (data.result) {
                if (data.result.action) {
                  switch (data.result.action) {
                    case "Msgs": {
                      if(!this.read_last_page){
                        this.read_last_page=true;
                      }
                      var _chats = data.result.data;
                      if (_chats && _chats.length > 0) {
                        //   _chats.reverse();
                        for (let _chat of _chats) {
                          var _append = this.handlerSetMaxMinChatId(_chat.id);
                          if (_append) {
                            this.msglist.push(_chat);
                          } else {
                            this.msglist.unshift(_chat);
                          }
                        }
                        //   this.msglist = [..._chats, ...this.msglist];
                        //推送消息出去，前面插入
                        this.handlerShareMessage('Msgs', _chats);
                      } else {
                        //没有更早的记录了
                        this.can_request_pre_chat_page = false;
                        this.handlerShareMessage('nodata', false);
                      }
                    }
                      break;
                    case "News": {
                      var _chats = data.result.data;
                      if (_chats && _chats.length > 0) {
                        for (let _chat of _chats) {
                          this.handlerSetMaxMinChatId(_chat.id);
                          //推送消息出去，前面插入
                          this.handlerShareChatMessage(_chat, true);
                        }
                        this.msglist.push(..._chats);
                      }
                    }
                      break;
                    case "Speak": {
                      if (data.result.code == 200) {
                        var _chat = data.result.data;
                        if (_chat) {
                          if (_chat.id) {
                            this.handlerSetMaxMinChatId(_chat.id);
                          }
                        }
                        this.msglist.push(_chat);
                        // if (_chat.msg_type == 10) {
                        //   return
                        // }
                        //推送到UI中
                        this.handlerShareChatMessage(_chat, true, 'Speak')
                      } else {
                        //data.result.msg;//
                        // this.handlerShareChatMessage(_chat,true,'Notice')
                        wx.showToast({
                          title: data.result.data,
                          icon: 'none'
                        })
                      }
                    }
                      break;
                    case "Join": {

                      var _join = data.result.data;

                      if(_join){
                        this.basicinfo ={
                          app_name:_join.app_name,
                          app_head:_join.app_head,
                          app_buttons:_join.buttons,
                          app_hidden:_join.app_hidden
                        };
                      }

                      if (_join?.queue_index) {
                        this.handlerShareMessage('Join', _join);
                      }
                      if (_join?.close_time) {
                        this.app.close_time = _join.close_time * 1000
                      }
                      if (_join) {
                        if (_join.visitor_token) {
                          wx.setStorageSync('visitor_token', _join.visitor_token)
                        }
                        if (_join.first) {
                          //首次链接，或者是短线重连
                          if (!_join.new_visitor) {
                            this.can_request_pre_chat_page = true
                            //当前是老访客，这里未完待续
                            if (this.addcount == 1) {
                              //获取历史记录 首次链接，则读取最后一页数据
                              this.privateGetSocketPrePageMessage();
                            }
                          }else{
                            this.read_last_page=true;
                          }
                        }
                        if (this.addcount > 1) {
                          //获取断层消息
                          this.handlerGetSocketFaultMessage();
                        }
                      }
                    }
                      break;
                    case "Message":
                    case "Score": {
                      this.handlerShareMessage(data.result.action, '');
                      wx.showToast({
                        title: '反馈成功，感谢您的参与!',
                        icon: 'none'
                      })
                    }
                      break;
                    case "TapWindow":
                    case "Cevent":
                      break;
                    default:
                      console.log('unhandler response action ', data.result.action, "data.type==3");
                      break;
                  }
                }
              }
            }
            break;
          case 6: {
          }
            break;
          default: {
            console.log('unhandler datatype ', data.type);
          }
            break;
        }
      }
    } catch (error) {
      console.error('消息解析错误', error);
    }
  }

  /**
   * 设置最小，最大ID
   * @param {*} chatid 
   */
  handlerSetMaxMinChatId(chatid) {
    var _append = true;
    if (this.chat_min_id > chatid || this.chat_min_id == 0) {
      this.chat_min_id = chatid;
      _append = false;
    }
    if (this.chat_max_id < chatid || this.chat_max_id == 0) {
      this.chat_max_id = chatid;
    }
    return _append;
  }
  /**
   * 获取所有历史记录
   */
  handlerChatMessages() {
    return this.msglist;
  }
/**
 * 开始心跳包
 */
  startHeartbeat() {
    if (this.heartbeatTimeout) clearInterval(this.heartbeatTimeout);
    this.heartbeatTimeout = setInterval(() => {
      this.sendMessage({
        type: 6
      });
      if (this.socket?.readyState == 3) {
        this.reconnect()
      }
    }, this.heartbeatInterval);
  }
/**
 * 准备重新链接，如果可以的话
 */
  reconnect() {
    if (this.reconnectTimer) clearTimeout(this.reconnectTimer);
    this.reconnectTimer = setTimeout(() => {
      this.initWebSocket();
    }, this.reconnectInterval);
  }
  /**
   * 在onHide的3秒后系统会断开websocket，所以在onShow的时候要检查下，否则onShow之后，有一段的空白期
   */
  checkLink() {
    if (this.socket) {
      if (this.socket.readyState == 3) {
        //重新链接
        this.initWebSocket();
      }
    } else {
      //重新连接
      this.initWebSocket();
    }
  }
  close() {
  }
}
module.exports = WebSocketManager;